@@ -188,16 +188,24 @@ let row_data stmt = Array.init (data_count stmt) (column stmt)
 let row_names stmt = Array.init (data_count stmt) (column_name stmt)
 let row_decltypes stmt = Array.init (data_count stmt) (column_decltype stmt)
 
-let exec_sql db sql f =
-  let rec loop stmt =
-    if step stmt = Rc.ROW then (
-      f stmt;
-      match prepare_tail stmt with
-      | Some stmt -> loop stmt
-      | None -> ())
-    else ()
-  in
-  loop (prepare db sql)
+
+(* Function registration *)
+
+external create_function :
+  db -> string -> int -> (Data.t array -> Data.t) -> unit =
+  "caml_sqlite3_create_function"
+
+let create_funN db name f = create_function db name (-1) f
+let create_fun0 db name f = create_function db name 0 (fun _ -> f ())
+let create_fun1 db name f = create_function db name 1 (fun args -> f args.(0))
+
+let create_fun2 db name f =
+  create_function db name 2 (fun args -> f args.(0) args.(1))
+
+let create_fun3 db name f =
+  create_function db name 3 (fun args -> f args.(0) args.(1) args.(2))
+
+external delete_function : db -> string -> unit = "caml_sqlite3_delete_function"
 
 
 (* Initialisation *)
